import numpy as np
import numpy.linalg as la
import scipy.linalg as sla
import scipy.linalg.lapack as lapack
from numpy.polynomial.legendre import leggauss as lgg
import scipy.optimize as opt

from neml import arbbar
from neml.math import nemlmath

from functools import partial

def generate_thickness_gradient(ri, ro, T1, T2, Tdot_hot, hold, 
    Tdot_cold = None, hold_together = 0.0, delay = 0.0):
  """
    Generate our usual through-wall temperature gradient

    Parameters:
      ri                inner radius
      ro                outer radius
      T1                initial temperature
      T2                hot temperature
      Tdot_hot          heating rate, also cooling rate if not specified
      hold              hold at gradient

    Optional:
      Tdot_cold         cooling rate
      hold_together     hold at no temperature gradient
  """
  if Tdot_cold is None:
    Tdot_cold = Tdot_hot

  dT = T2 - T1
  thot = np.abs(dT / Tdot_hot)
  tcold = np.abs(dT / Tdot_cold)
  period = thot + hold + tcold + hold_together

  def gradient(r, t):
    if t < delay:
      return T1
    t = (t-delay) % period

    if t < thot:
      Tright = T1 + t * Tdot_hot * np.sign(dT)
    elif t < thot + hold:
      Tright = T2
    elif t < thot + hold + tcold:
      Tright = T2 - (t - thot - hold) * Tdot_cold * np.sign(dT)
    else:
      Tright = T1

    xi = (r - ri) / (ro - ri)

    return (1-xi) * T1 + xi * Tright

  return gradient

def generate_standard_timesteps(T1, T2, Tdot_hot, hold, nheat, nhold,
    ncycles, 
    Tdot_cold = None, ncool = None, hold_together = 0.0,
    ntogether = None, delay = 0.0, ndelay = None):
  """
    Generate a standard time stepping scheme for the thickness gradients

    Parameters:
      T1                initial temperature
      T2                hot temperature
      Tdot_hot          heat rate, also cooling rate if not specified
      hold              hold at gradient
      nheat             number of steps during heating
      nhold             number of steps during hold
      ncycles           number of cycles

    Optional:
      Tdot_cold         cooling rate
      ncool             number of cooling steps
      hold_together     hold at zero gradient
      ntogether         number of iso hold steps
      delay             delay before cycling
      ndelay            number of delay steps
  """
  if Tdot_cold is None:
    Tdot_cold = Tdot_hot
  if ncool is None:
    ncool = nheat

  if (delay != 0.0) and (ndelay is None):
    raise ValueError("If the delay is not zero you need to provide ndelay!")

  if (hold_together != 0.0) and (ntogether is None):
    raise ValueError("If you provide hold_together you need to provide ntogether!")

  times = [0.0]
  if delay > 0.0:
    times.extend(list(times[-1] + np.linspace(0, delay, ndelay+1)[1:]))

  for i in range(ncycles):
    times.extend(list(times[-1] + np.linspace(0, np.abs(T2-T1) / Tdot_hot, 
      nheat+1)[1:]))
    if hold > 0.0:
      times.extend(list(times[-1] + np.linspace(0, hold, nhold+1)[1:]))
    times.extend(list(times[-1] + np.linspace(0, np.abs(T2-T1) / Tdot_cold,
      ncool+1)[1:]))
    if hold_together > 0.0:
      times.extend(list(times[-1] + np.linspace(0, hold_together, 
        ntogether+1)[1:]))

  return times

def generate_multimaterial_thickness_gradient(rs, ks, T1, T2, Tdot_hot,
    hold, Tdot_cold = None, hold_together = 0.0, delay = 0.0, flip = False):
  """
    Generate a thermal gradient alternating between hot and cold for the
    case of a multilayer material.

    This function assumes the "slab" limit, i.e. a thin walled section.

    Parameters:
      rs                radii
      ks                heat conduction coefficients
      T1                initial temperature
      T2                hot temperature
      Tdot_hot          heating rate, also cooling rate if not specified
      hold              hold at gradient

    Optional:
      Tdot_cold         cooling rate
      hold_together     hold at no temperature gradient
      flip              have gradient going the other direction
  """
  if Tdot_cold is None:
    Tdot_cold = Tdot_hot

  if len(rs) != (len(ks)+1):
    raise ValueError("Incompatible radii and thermal conductivity arrays")

  dT = T2 - T1
  thot = np.abs(dT / Tdot_hot)
  tcold = np.abs(dT / Tdot_cold)
  period = thot + hold + tcold + hold_together

  def gradient(r, t):
    if t < delay:
      return T1
    t = (t-delay) % period

    if t < thot:
      Tright = T1 + t * Tdot_hot * np.sign(dT)
    elif t < thot + hold:
      Tright = T2
    elif t < thot + hold + tcold:
      Tright = T2 - (t - thot - hold) * Tdot_cold * np.sign(dT)
    else:
      Tright = T1

    Rs = np.diff(rs) / np.array(ks)
    Rt = np.sum(Rs)
    
    if flip:
      Ti = list(T1 + (1-np.cumsum(np.insert(Rs,0,0.0)) / Rt) * (Tright - T1)) 
    else:
      Ti = list(T1 + np.cumsum(np.insert(Rs,0,0.0)) / Rt * (Tright - T1))

    return np.interp(r, rs, Ti)

  return gradient

def mesh(rs, ns, bias = False, n = 2.0):
  """
    Generate a 1D mesh

    Parameters
      rs        list of points which divide the line into segments
                each segment is guaranteed to be a node point
                
                x----------x-------x--------------x
                r0         r1      r2             r3

      ns        list giving the number of elements in each segment

    Optional:
      bias      bias the mesh towards the segment end points
      n         bias factor
  """
  if not bias:
    xpoints = [np.linspace(rs[0], rs[1], ns[0]+1)]
    for i in range(1, len(ns)):
      xpoints.append(np.linspace(rs[i], rs[i+1], ns[i]+1)[1:])
    xpoints = np.concatenate(tuple(xpoints))
  else:
    xpoints = [rs[0]]
    for i in range(len(rs)-1):
      if ns[i] % 2 == 0:
        even = True
        nuse = ns[i] // 2 + 1
      else:
        even = False
        nuse = (ns[i] + 1) // 2 + 1

      x1 = (np.linspace(0,1, nuse)[1:])**n
      x2 = 1-(np.linspace(0,1, nuse)[:-1])**n + 1.0
      if even:
        xss = np.concatenate((x1, np.flip(x2,0)))/2
      else:
        xss = np.concatenate((x1[:-1], np.flip(x2,0)))/2
      
      # Transform
      xss = xss * (rs[i+1] - rs[i]) + rs[i]
      xpoints.extend(list(xss))
    
    xpoints = np.array(xpoints)

  return xpoints

class VesselSectionProblem(object):
  """
    This serves as the superclass for both the models of a "section of a
    cylindrical vessel": the Bree problem and my steady-state axisymmetric 
    model.
  """
  def __init__(self, rs, mats, ns, T, p, rtol = 1.0e-6, atol = 1.0e-8,
      p_ext = lambda t: 0.0):
    """
      Parameters:
        rs      radii deliminating each region
        mats    material model for each region
        ns      number of elements per region
        T       temperature as a function of (r, t)
        p       pressure, as a function of (t)

      Optional:
        rtol    relative tolerance for N-R solve
        atol    absolute tolerance for N-R solve
        p_ext   external pressure, as a function of t
    """
    # Check
    if len(rs) - 1 != len(mats):
      raise ValueError("Length of radii must be one more then length of regions")
    if len(mats) != len(ns):
      raise ValueError("Inconsistent region definitions")

    self.rs = rs
    self.mats = mats
    self.ns = ns
    self.T = T
    self.p = p
    self.p_ext = p_ext

    self.rtol = rtol
    self.atol = atol

    self.ts = np.diff(rs)
    self.t = np.sum(self.ts)
    self.r = rs[-1]
    self.r_inner = rs[0]
    self.r_outer = rs[-1]

    self.times = [0.0]
    self.pressures = [self.p(0)]
    self.external_pressures = [self.p_ext(0.0)]

def rect(n):
  """
    Rectangle rule points and weights over the
    standard domain [-1,1]

    Parameters:
      n     number of points
  """
  npts = np.linspace(-1,1,n+1)
  pts = np.array([(a+b)/2 for a,b in zip(npts[0:],npts[1:])])
  weights = np.array([2.0/n]*len(pts))

  return pts, weights

class BreeProblem(VesselSectionProblem):
  """
    A Bree problem, referenced back to a vessel.

    If you'd rather have a "standard" Bree problem you can
    make the overall wall thickness = 1 spanning from
    r = [0,1] and directly apply the hoop stress as the 
    pressure.
  """
  def __init__(self, rs, mats, ns, T, p, rtol = 1.0e-6, atol = 1.0e-8,
      itype = "gauss", p_ext = lambda t: 0.0):
    """
      Parameters:
        rs      radii deliminating each region
        mats    material model for each region
        ns      number of elements per region
        T       temperature as a function of (r, t)
        p       pressure, as a function of (t)

      Optional:
        rtol    relative tolerance for N-R solve
        atol    absolute tolerance for N-R solve
        bias    bias the mesh towards the interfaces
        itype   "rectangle" or "gauss", defaults to "gauss"
        p_ext   external pressure as a function of t
    """
    super(BreeProblem, self).__init__(rs, mats, ns, T, p,
        rtol = rtol, atol = atol, p_ext = p_ext)
    
    self.regions = np.diff(self.rs)
    self.dlength = [r for r,n in zip(self.regions, self.ns) for i in range(n)]

    if itype == "gauss":
      ifn = lgg
    elif itype == "rectangle":
      ifn = rect
    else:
      raise ValueError("Unknown integration type %s!" % itype)

    self.ipoints = [(xi+1)*dl/2 + r for dl,r,n in zip(self.regions, self.rs, self.ns) for 
        xi in ifn(n)[0]]
    self.weights = [wi * dl/2 for dl,n in zip(self.regions, self.ns) for 
        wi in ifn(n)[1]]
    materials = [m for m,n in zip(self.mats, self.ns) for i in range(n)]

    self.npoints = len(self.ipoints)
    
    self.P = lambda t: (p(t) * self.r_inner - p_ext(t) * self.r_outer) / self.t

    # Setup the n bar model
    self.barmodel = arbbar.BarModel()
    self.barmodel.add_node(1)
    self.barmodel.add_node(2)

    for ipt, wt, mat in zip(self.ipoints, self.weights, materials):
      def tlocal(tt, x = ipt):
        return self.T(x, tt)
      self.barmodel.add_edge(1,2, object = arbbar.Bar(mat, wt, 1.0, 
        T = tlocal))

    self.barmodel.add_displacement_bc(1, lambda t: 0.0)
    self.barmodel.add_force_bc(2, lambda t: self.P(t) * np.sum(self.weights))

    # Setup fields
    self.axialstrain = [0.0]
    self.hoop = [self.P(0.0)]
    self.force = [0.0]
    self.temperatures = [np.array([self.barmodel[1][2][i]['object'].temperature[-1]
      for i in range(self.npoints)])]
    self.stresses = [np.array([self.barmodel[1][2][i]['object'].stress[-1]
      for i in range(self.npoints)])]
    self.tstrains = [np.array([self.barmodel[1][2][i]['object'].tstrain[-1]
      for i in range(self.npoints)])]
    self.mstrains = [np.array([self.barmodel[1][2][i]['object'].mstrain[-1]
      for i in range(self.npoints)])]
    self.estrains = [np.array([self.barmodel[1][2][i]['object'].estrain[-1]
      for i in range(self.npoints)])]
    self.histories = [np.array([self.barmodel[1][2][i]['object'].history[-1]
      for i in range(self.npoints)])]

    self.energy = [0.0]
    self.work = [0.0]

  def update_loading(self, new_p, new_T, new_p_ext = lambda t: 0.0):
    """
      Impose new loading functions
    """
    self.P = lambda t: (new_p(t) * self.r_inner - new_p_ext(t) * self.r_outer) / self.t
    self.barmodel.nodes[2]['force bc'] = lambda t: self.P(t) * np.sum(self.weights)

    self.T = new_T
    for i,ipt in enumerate(self.ipoints):
      def tlocal(tt, x = ipt):
        return self.T(x, tt) 
      self.barmodel[1][2][i]['object'].T = tlocal

  def step(self, t, rtol = 1.0e-6, atol = 1.0e-8, verbose = False,
      ndiv = 4, dfact = 2, extrapolate = False):
    """
      Advance to the next step

      Parameters:
        t           time requested

      Optional:
        rtol        solver relative tolerance
        atol        solver absolute tolerance
        verbose     print a lot of stuff
        ndiv        maximum number of adaptive subdivisions
        dfact       subdivision factor
        extrapolate try to extrapolate displacements
    """
    dt = t - self.times[-1]
    if dt < 0.0:
      raise ValueError("Requesting negative time step!")
    
    self.barmodel.solve(dt, ndiv = ndiv, dfact = dfact, 
        verbose = verbose, extrapolate = extrapolate)
    
    self.times.append(t)
    self.pressures.append(self.p(t))
    self.external_pressures.append(self.p_ext(t))
    self.hoop.append(self.P(t))
    self.force.append(self.barmodel.nodes[2]['forces'][-1])
    self.axialstrain.append(self.barmodel.nodes[2]['displacements'][-1])
    
    self.temperatures.append(np.array([self.barmodel[1][2][i]['object'].temperature[-1]
      for i in range(self.npoints)]))
    self.stresses.append(np.array([self.barmodel[1][2][i]['object'].stress[-1]
      for i in range(self.npoints)]))
    self.tstrains.append(np.array([self.barmodel[1][2][i]['object'].tstrain[-1]
      for i in range(self.npoints)]))
    self.mstrains.append(np.array([self.barmodel[1][2][i]['object'].mstrain[-1]
      for i in range(self.npoints)]))
    self.estrains.append(np.array([self.barmodel[1][2][i]['object'].estrain[-1]
      for i in range(self.npoints)]))
    self.histories.append(np.array([self.barmodel[1][2][i]['object'].history[-1]
      for i in range(self.npoints)]))

    self.energy.append(0.0)
    self.work.append(0.0)
    for i,wt in enumerate(self.weights):
      me = self.barmodel[1][2][i]['object']
      self.energy[-1] += wt*me.energy[-1]*me.A*me.l
      self.work[-1] += wt*me.dissipation[-1]*me.A*me.l
    
class AxisymmetricProblem(VesselSectionProblem):
  """
    Python driver for a axisymmetric reduced problem.

    My coordinate system uses:
    x = radius
    y = theta
    z = z

    when we call into the material model.

    The conditions here are "steady state axisymmetry."
    The model is 1D and represents the stress state away from structural 
    discontinuities.

    The default boundary conditions are axisymmetry + generalized plane strain.
    The generalized plane strain boundary condition prevents initially
    parallel cross sections from rotating, but allows some net axial strain.

    Using constrained = True gives traditional plane strain
  """
  def __init__(self, rs, mats, ns, T, p, rtol = 1.0e-6, atol = 1.0e-8,
      bias = False, factor = 2.0, ngpts = 1, constrained = False,
      p_ext = lambda t: 0.0):
    """
      Parameters:
        rs      radii deliminating each region
        mats    material model for each region
        ns      number of elements per region
        T       temperature as a function of (r, t)
        p       pressure, as a function of (t)

      Optional:
        rtol        relative tolerance for N-R solve
        atol        absolute tolerance for N-R solve
        bias        bias the mesh towards the interfaces
        factor      bias factor
        ngpts       number of gauss points per element
        constrained constrain the cylinder against thermal expansion
        p_ext       external pressure, as a function of time
        constrained fully constrain the cylinder against thermal expansion
    """
    super(AxisymmetricProblem, self).__init__(rs, mats, ns, T, p,
        rtol = rtol, atol = atol, p_ext = p_ext)

    self.constrained = constrained

    self.mesh = mesh(self.rs, self.ns, bias = bias, n = factor)
    self.nnodes = len(self.mesh)
    self.nelem = len(self.mesh) - 1
    self.ls = np.diff(self.mesh)

    self.materials = [m for m,n in zip(self.mats, self.ns) for i in range(n)]
    
    self.displacements = [[0.0 for i in range(self.nnodes)]]
    self.axialstrain = [0.0]
    
    self.ngpts = ngpts
    self.gpoints, self.gweights = lgg(ngpts)

    self.temperatures = [np.array([self.T(xi, 0) for xi in self.mesh])]

    self.histories = [[np.array([m.init_store() for xi in self.gpoints])
      for m in self.materials]]
    self.stresses = [np.array([np.zeros((self.ngpts,6)) for n in self.ns for i in range(n)])]
    self.strains = [np.array([np.zeros((self.ngpts,6)) for n in self.ns for i in range(n)])]
    self.tstrains = [np.array([np.zeros((self.ngpts,6)) for n in self.ns for i in range(n)])]
    self.mstrains = [np.array([np.zeros((self.ngpts,6)) for n in self.ns for i in range(n)])]
    self.estrains = [np.array([np.zeros((self.ngpts,6)) for n in self.ns for i in range(n)])]

    # Save a bit of time
    self.Nl = np.array([[(1-xi)/2, (1+xi)/2] for xi in self.gpoints])
    self.ri = np.array([np.dot(self.Nl, self.mesh[e:e+2]) for e in range(self.nelem)])
    self.Bl = np.array([[-1.0/2, 1.0/2] for xi in self.gpoints])

  def take_step(self, t, rtol = 1.0e-6, atol = 1.0e-8, ilimit = 10,
      verbose = False, predict = 1.0):
    """
      Actually take a time step, used to sort out adaptive integration
      
      Parameters:
        t       next time

      Optional:
        rtol    solver relative tolerance
        atol    solver absolute tolerance
        ilimit  solver iteration limit
        verbose flag to print a lot of convergence info
        predict factor for forward extrapolation for the initial guess:
                1.0 = full extrapolation
                0.0 = use previous step
    """
    T = np.array([self.T(xi, t) for xi in self.mesh])
    p_left = self.p(t)
    p_right = self.p_ext(t)

    # Get a guess
    if len(self.displacements) < 2:
      x0 = np.concatenate((self.displacements[-1] ,
        [self.axialstrain[-1]]))
    else:
      x0 = np.concatenate((self.displacements[-1] + (self.displacements[-1] - self.displacements[-2]) * predict,
        [self.axialstrain[-1] + (self.axialstrain[-1] - self.axialstrain[-2]) * predict]))

    x = np.copy(x0)
    if self.constrained:
      x[-1] = 0.0

    R, (J11_du,J11_d,J11_dl,J12,J21,J22), strains, tstrains, mstrains, estrains, stresses, histories = self.RJ(x, T, p_left,
        p_right, t)

    if self.constrained:
      nR0 = la.norm(R[:-1])
    else:
      nR0 = la.norm(R)
    nR = nR0

    if verbose:
      print("Iter.\tNorm res")
      print("%i\t%e" % (0, nR0))
    
    i = 0
    while (nR > atol) and (nR / nR0 > rtol) and (i < ilimit):
      if self.constrained:
        R1 = R[:-1]
        x1 = tri_solve(J11_du, J11_d, J11_dl, R1)
        x[:-1] -= x1

      else:
        R1 = R[:-1]
        R2 = R[-1]
        c = R1 - R2/J22 * J12
        u = -J12 / J22
        v = J21
        x1 = cute_solve(J11_du, J11_d, J11_dl, u, v, c)
        x2 = (R2 - np.dot(J21,x1)) / J22

        x[:-1] -= x1
        x[-1] -= x2

      R, (J11_du,J11_d,J11_dl,J12,J21,J22), strains, tstrains, mstrains, estrains, stresses, histories = self.RJ(x, T, p_left,
          p_right, t)

      if self.constrained:
        nR = la.norm(R[:-1])
      else:
        nR = la.norm(R)
      i += 1

      if verbose:
        print("%i\t%e" % (i, nR))

    if i == ilimit:
      raise RuntimeError("Exceeded maximum allowed iterations!")
    
    return p_left, p_right, T, strains, tstrains, mstrains, estrains, stresses, histories, x[:-1], x[-1]

  def step(self, t, rtol = 1.0e-6, atol = 1.0e-8, verbose = False,
      div = 2, max_div = 4, predict = 1.0, ilimit = 10):
    """
      Update to the next state

      Parameters:
        t       time requested

      Optional:
        rtol    solver relative tolerance
        atol    solver absolute tolerance
        verbose flag to print a lot of convergence info
        div     step division factor for adaptive stepping: 
                new_time_step = old_time_step / div
        max_div maximum number of adaptive subdivisions
        predict factor for forward extrapolation for the initial guess:
                1.0 = full extrapolation
                0.0 = use previous step
        ilimit  solver iteration limit
    """
    istep = div**max_div
    sstep = istep
    cstep = 0
    cdiv = 0
    dt = t - self.times[-1]

    ostep = len(self.times)

    while cstep != istep:
      cstep += sstep      
      try:
        p_left, p_right, T, strains, tstrains, mstrains, estrains, stresses, histories, disp, axial = self.take_step(
            self.times[-1] + dt * float(cstep)/istep, rtol = rtol, atol = atol,
            verbose = verbose, predict = predict, ilimit = ilimit)
        self.times.append(t)
        self.pressures.append(p_left)
        self.external_pressures.append(p_right)
        self.temperatures.append(T)
        self.strains.append(strains)
        self.tstrains.append(tstrains)
        self.mstrains.append(mstrains)
        self.estrains.append(estrains)
        self.stresses.append(stresses)
        self.histories.append(histories)
        self.displacements.append(disp)
        self.axialstrain.append(axial)
      except RuntimeError as e:
        cstep -= sstep
        cdiv += 1
        if cdiv > max_div:
          self.times = self.times[:ostep]
          self.pressures = self.pressures[:ostep]
          self.external_pressures = self.external_pressures[:ostep]
          self.temperatures = self.temperatures[:ostep]
          self.strains = self.strains[:ostep]
          self.tstrains = self.tstrains[:ostep]
          self.mstrains = self.mstrains[:ostep]
          self.estrains = self.estrains[:ostep]
          self.stresses = self.stresses[:ostep]
          self.histories = self.histories[:ostep]
          self.displacements = self.displacements[:ostep]
          self.axialstrain = self.axialstrain[:ostep]
          raise e
        sstep /= div
        if verbose:
          print("Substepping: new step fraction %f" % (float(sstep) / istep))

    self.times = self.times[:ostep] + [self.times[-1]]
    self.pressures = self.pressures[:ostep] + [self.pressures[-1]]
    self.external_pressures = self.external_pressures[:ostep] + [self.external_pressures[-1]]
    self.temperatures = self.temperatures[:ostep] + [self.temperatures[-1]]
    self.strains = self.strains[:ostep] + [self.strains[-1]]
    self.tstrains = self.tstrains[:ostep] + [self.tstrains[-1]]
    self.mstrains = self.mstrains[:ostep] + [self.mstrains[-1]]
    self.estrains = self.estrains[:ostep] + [self.estrains[-1]]
    self.stresses = self.stresses[:ostep] + [self.stresses[-1]]
    self.histories = self.histories[:ostep] + [self.histories[-1]]
    self.displacements = self.displacements[:ostep] + [self.displacements[-1]]
    self.axialstrain = self.axialstrain[:ostep] + [self.axialstrain[-1]]


  def strain(self, d, l, r, ez):
    """
      Compute the full strains

      Parameters:
        d   displacements
        l   length
        r   radius
        ez  axial strain
    """
    strain = np.zeros((len(r),6))
    strain[:,0] = np.dot(self.Bl, d) * 2.0 / l
    strain[:,1] = np.dot(self.Nl, d) / r
    strain[:,2] = ez 
    return strain

  def RJ(self, x, T, p_left, p_right, t):
    """
      Compute the residual, jacobian, and the rest of the updated quantities

      Parameters:
        x           iterate
        T           nodal temperatures
        Tn          previous temperatures
        p_left      left pressure
        t           time
        p_right     right pressure
        t           time 
    """
    d = x[:-1]
    ez = x[-1]

    Fext = np.zeros((self.nnodes+1,))
    Fext[0] = p_left
    Fext[-2] = -p_right
    #Fext[-1] = (p_left *self.r_inner - p_right * self.r_outer) / (2*self.t)
    # We should use the proper thick-walled vessel formula
    Fext[-1] = (p_left * self.r_inner**2.0 / (self.r_outer**2.0 - self.r_inner**2.0) 
        -p_right * self.r_outer**2.0 / (self.r_outer**2.0 - self.r_inner**2.0))
    
    strains = np.zeros((self.nelem, self.ngpts, 6))
    tstrains = np.zeros((self.nelem, self.ngpts, 6))
    mstrains = np.zeros((self.nelem, self.ngpts, 6))
    estrains = np.zeros((self.nelem, self.ngpts, 6))
    stresses = np.zeros((self.nelem, self.ngpts, 6))
    histories = []

    Fint = np.zeros((self.nnodes+1,))
    A11_dl = np.zeros((self.nnodes-1,))
    A11_d = np.zeros((self.nnodes,))
    A11_du = np.zeros((self.nnodes-1,))
    A12 = np.zeros((self.nnodes,))
    A21 = np.zeros((self.nnodes,))
    A22 = 0.0

    for e, (mat, l, rs) in enumerate(zip(self.materials, self.ls, self.ri)):
      de = d[e:e+2]
      T_e = np.dot(self.Nl, T[e:e+2])
      T_n = np.dot(self.Nl, self.temperatures[-1][e:e+2])
      
      therm_inc = np.array([mat.alpha(Ti)*(Ti-Tii)*np.array([1,1,1,0,0,0]) 
        for Ti, Tii in zip(T_e, T_n)])

      strain = self.strain(de, l, rs, ez)
      tstrain = self.tstrains[-1][e] + therm_inc
      mstrain = strain - tstrain
      mstrain_n = self.mstrains[-1][e]
      stress_n = self.stresses[-1][e]
      hist_n = self.histories[-1][e]

      stress = np.zeros((self.ngpts,6))
      history = np.zeros((self.ngpts,len(hist_n[0])))
      estrain = np.zeros((self.ngpts,6))

      for i in range(self.ngpts):
        si, hi, ti, ui, pi = mat.update_sd(mstrain[i], mstrain_n[i],
            T_e[i], T_n[i], t, self.times[-1], stress_n[i], hist_n[i], 
            0.0, 0.0)
        stress[i] = si
        history[i] = hi

        estrain[i] = mat.elastic_strains(si, T_e[i], hi)
        
        wi = self.gweights[i]
        #Actually add in our terms
        Fint[e:e+2] += wi * (si[0] * self.Bl[i] * 2.0 / l + (si[1] - si[0]) * self.Nl[i] / rs[i]) * l / 2.0
        Fint[-1] += wi * si[2] * l / 2.0 / self.t
        
        DE = np.array([self.Bl[i] * 2.0/l, self.Nl[i] / rs[i]])
        
        # Remember wi and the jacobian
        J11 = np.dot(np.outer(2.0/l * self.Bl[i], ti[0,:2]) + np.outer(self.Nl[i] / rs[i], 
            ti[1,:2] - ti[0,:2]), DE) * wi * l / 2.0 
        J12 = (ti[0,2] * self.Bl[i] * 2.0 / l + (ti[1,2] - ti[0,2]) * self.Nl[i] / rs[i]) * wi * l / 2.0
        J21 = (np.dot(ti[2,:2], DE) / self.t) * wi * l / 2.0 
        J22 = (ti[2,2] / self.t) * wi * l / 2.0

        A11_du[e] += J11[0,1]
        A11_d[e] += J11[0,0]
        A11_d[e+1] += J11[1,1]
        A11_dl[e] += J11[1,0]
        A12[e:e+2] += J12
        A21[e:e+2] += J21
        A22 += J22

      strains[e] = strain
      tstrains[e] = tstrain
      mstrains[e] = mstrain
      estrains[e] = estrain
      stresses[e] = stress
      histories.append(history)

    return Fint - Fext, (A11_du,A11_d,A11_dl,A12,A21,A22), strains, tstrains, mstrains, estrains, stresses, histories

def tri_solve(DU, D, DL, c):
  """
    Solve B x = c
  """
  DL, D, DU, DU2, IPIV = nemlmath.dgttrf(DL, D, DU)
  x = nemlmath.dgttrs(DL,D,DU,DU2,IPIV,c)

  return x

def cute_solve(DU, D, DL, u, v, c):
  """
    Do a cute solve of:

    (B + u v.T) x = c

    with B tri-diagonal
  """
  DL, D, DU, DU2, IPIV = nemlmath.dgttrf(DL, D, DU)
  
  x = nemlmath.dgttrs(DL,D,DU,DU2,IPIV, c)
  y = nemlmath.dgttrs(DL,D,DU,DU2,IPIV, u)
  
  return x - np.dot(v,x) / (1+np.dot(v,y)) * y



import sys
sys.path.append('..')

from neml import interpolate, solvers, models, elasticity, ri_flow, hardening, surfaces, visco_flow, general_flow, creep
from common import *

import unittest
import numpy as np
import numpy.linalg as la

class CommonMatModel(object):
  """
    Tests that could be applied to all material models
  """
  def test_tangent_proportional_strain(self):
    t_n = 0.0
    strain_n = np.zeros((6,))
    stress_n = np.zeros((6,))
    hist_n = self.model.init_store()

    u_n = 0.0
    p_n = 0.0

    for i,m in enumerate(np.linspace(0,1,self.nsteps)):
      t_np1 = self.tfinal * m
      strain_np1 = self.efinal * m
      
      stress_np1, hist_np1, A_np1, u_np1, p_np1 = self.model.update_sd(
          strain_np1, strain_n, self.T, self.T, t_np1, t_n, stress_n, hist_n,
          u_n, p_n)
      dfn = lambda e: self.model.update_sd(e,
          strain_n, self.T, self.T, t_np1, t_n, stress_n, hist_n, u_n, p_n)[0]
      num_A = differentiate(dfn, strain_np1, eps = 1.0e-9)
      
      if not np.allclose(num_A, A_np1, rtol = 1e-3, atol = 1e-1):
        print(A_np1)
        print(num_A)

      self.assertTrue(np.allclose(num_A, A_np1, rtol = 1.0e-3, atol = 1.0e-1))
      
      strain_n = strain_np1
      stress_n = stress_np1
      hist_n = hist_np1
      u_n = u_np1
      p_n = p_np1

class TestPerfectPlasticity(unittest.TestCase, CommonMatModel):
  def setUp(self):
    self.E = 92000.0
    self.nu = 0.3

    self.mu = self.E/(2*(1+self.nu))
    self.K = self.E/(3*(1-2*self.nu))

    self.s0 = 180.0

    self.elastic = elasticity.IsotropicLinearElasticModel(self.mu,
        "shear", self.K, "bulk")

    surface = surfaces.IsoJ2()
    self.model = models.SmallStrainPerfectPlasticity(
        self.elastic, surface, self.s0,
        max_divide = 5, force_divide = True)

    self.efinal = np.array([0.1,-0.05,0.02,-0.03,0.1,-0.15])
    self.tfinal = 10.0
    self.T = 300.0
    self.nsteps = 10

class TestRIAPlasticityJ2Linear(unittest.TestCase, CommonMatModel):
  """
    Test the rate-independent plasticity algorithm with a linearly
    isotropically hardening yield surface.
  """
  def setUp(self):
    self.hist0 = np.zeros((7,))

    self.E = 92000.0
    self.nu = 0.3

    self.mu = self.E/(2*(1+self.nu))
    self.K = self.E/(3*(1-2*self.nu))

    self.s0 = 180.0
    self.Kp = self.E/10

    self.elastic = elasticity.IsotropicLinearElasticModel(self.mu,
        "shear", self.K, "bulk")

    surface = surfaces.IsoJ2()
    hrule = hardening.LinearIsotropicHardeningRule(self.s0, self.Kp)
    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    self.model = models.SmallStrainRateIndependentPlasticity(self.elastic,
        flow, max_divide = 5, force_divide = True)

    self.efinal = np.array([0.1,-0.05,0.02,-0.03,0.1,-0.15])
    self.tfinal = 10.0
    self.T = 300.0
    self.nsteps = 10

class TestDirectIntegrateChaboche(unittest.TestCase, CommonMatModel):
  """
    Test Chaboche's VP model with our new direct integrator
  """
  def setUp(self):
    n = 20.0
    eta = 108.0
    sY = 89.0

    Q = 165.0
    b = 12.0
    
    self.m = 3

    C1 = 80.0e3
    C2 = 14.02e3
    C3 = 3.333e3

    y1 = 0.9e3
    y2 = 1.5e3
    y3 = 1.0

    surface = surfaces.IsoKinJ2()
    iso = hardening.VoceIsotropicHardeningRule(sY, Q, b)
    cs = [C1, C2, C3]
    gs = [y1, y2, y3]
    As = [0.0, 0.0, 0.0]
    ns = [1.0, 1.0, 1.0]
    gmodels = [hardening.ConstantGamma(g) for g in gs]
    hmodel = hardening.Chaboche(iso, cs, gmodels, As, ns)

    fluidity = visco_flow.ConstantFluidity(eta)

    self.hist0 = np.zeros((19,))
    self.T = 300.0

    vmodel = visco_flow.ChabocheFlowRule(surface, hmodel, fluidity, n)

    E = 92000.0
    nu = 0.3

    mu = E/(2*(1+nu))
    K = E/(3*(1-2*nu))

    self.elastic = elasticity.IsotropicLinearElasticModel(mu,
        "shear", K, "bulk")

    flow = general_flow.TVPFlowRule(self.elastic, vmodel)

    self.model = models.GeneralIntegrator(self.elastic, flow,
        max_divide = 3, force_divide = True)

    self.efinal = np.array([0.05,0,0,0.02,0,-0.01])
    self.tfinal = 10.0
    self.T = 300.0
    self.nsteps = 100



#!/usr/bin/env python3

from neml import models, interpolate, elasticity, history
from neml.cp import crystallography, slipharden, sliprules, inelasticity, kinematics, singlecrystal, batch
from neml.math import rotations, tensors

import unittest
import numpy as np

class TestBatch(unittest.TestCase):
  def setUp(self):
    self.N = 10

    self.tau0 = 10.0
    self.tau_sat = 50.0
    self.b = 2.5

    self.strengthmodel = slipharden.VoceSlipHardening(self.tau_sat, self.b, self.tau0)
    
    self.g0 = 1.0
    self.n = 3.0
    self.slipmodel = sliprules.PowerLawSlipRule(self.strengthmodel, self.g0, self.n)

    self.imodel = inelasticity.AsaroInelasticity(self.slipmodel)

    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")

    self.mu = 29000.0
    self.E = 120000.0
    self.nu = 0.3

    self.emodel = elasticity.CubicLinearElasticModel(self.E, 
        self.nu, self.mu, "moduli")

    self.kmodel = kinematics.StandardKinematicModel(self.emodel, self.imodel)

    self.model = singlecrystal.SingleCrystalModel(self.kmodel, self.L, 
        miter = 120)

    self.orientations = rotations.random_orientations(self.N)

    self.D = np.array([0.01,-0.002,-0.003,0.012,-0.04,0.01])
    self.W = np.array([0.02,-0.02,0.03])

    self.T = 300.0
    self.dt = 2.0

  def test_batch_init(self):
    H1 = batch.init_history_batch(self.model, self.N)
    H2 = np.array([self.model.init_store() for i in range(self.N)])

    self.assertTrue(np.allclose(H1, H2))

  def test_batch_set_get(self):
    H = batch.init_history_batch(self.model, self.N)

    batch.set_orientation_passive_batch(self.model, H, self.orientations)

    nq = batch.get_orientation_passive_batch(self.model, H)

    for q1,q2 in zip(self.orientations,nq):
      self.assertTrue(np.allclose(q1.quat,q2.quat))

  def test_batch_no_threads(self):
    self.batch_run(1)

  def test_batch_threads(self):
    self.batch_run(2)

  def batch_run(self, nthreads):
    h_n = batch.init_history_batch(self.model, self.N)
    batch.set_orientation_passive_batch(self.model, h_n, self.orientations)

    d_np1 = np.array([self.D for i in range(self.N)])
    w_np1 = np.array([self.W for i in range(self.N)])

    T_np1 = np.array([self.T for i in range(self.N)])
    T_n = np.array([self.T for i in range(self.N)])

    d_n = np.zeros((self.N,6))
    w_n = np.zeros((self.N,3))

    s_n = np.zeros((self.N,6))
    u_n = np.zeros((self.N,))
    p_n = np.zeros((self.N,))

    s_np1, h_np1, A_np1, B_np1, u_np1, p_np1 = batch.evaluate_crystal_batch(
        self.model, d_np1, d_n, w_np1, w_n, T_np1, T_n, self.dt, 0.0,
        s_n, h_n, u_n, p_n, nthreads = nthreads)

    other_s_np1 = np.array([self.model.update_ld_inc(
      d_np1_i, d_n_i, w_np1_i, w_n_i, T_np1_i, T_n_i, self.dt, 0.0,
      s_n_i, h_n_i, u_n_i, p_n_i)[0] for d_np1_i, d_n_i, w_np1_i,
      w_n_i, T_np1_i, T_n_i, s_n_i, h_n_i, u_n_i, p_n_i in zip(
        d_np1, d_n, w_np1, w_n, T_np1, T_n, s_n, h_n, u_n, p_n)])

    self.assertTrue(np.allclose(s_np1, other_s_np1))




import pytest
import re

from pybind11_tests import factory_constructors as m
from pybind11_tests.factory_constructors import tag
from pybind11_tests import ConstructorStats


def test_init_factory_basic():
    """Tests py::init_factory() wrapper around various ways of returning the object"""

    cstats = [ConstructorStats.get(c) for c in [m.TestFactory1, m.TestFactory2, m.TestFactory3]]
    cstats[0].alive()  # force gc
    n_inst = ConstructorStats.detail_reg_inst()

    x1 = m.TestFactory1(tag.unique_ptr, 3)
    assert x1.value == "3"
    y1 = m.TestFactory1(tag.pointer)
    assert y1.value == "(empty)"
    z1 = m.TestFactory1("hi!")
    assert z1.value == "hi!"

    assert ConstructorStats.detail_reg_inst() == n_inst + 3

    x2 = m.TestFactory2(tag.move)
    assert x2.value == "(empty2)"
    y2 = m.TestFactory2(tag.pointer, 7)
    assert y2.value == "7"
    z2 = m.TestFactory2(tag.unique_ptr, "hi again")
    assert z2.value == "hi again"

    assert ConstructorStats.detail_reg_inst() == n_inst + 6

    x3 = m.TestFactory3(tag.shared_ptr)
    assert x3.value == "(empty3)"
    y3 = m.TestFactory3(tag.pointer, 42)
    assert y3.value == "42"
    z3 = m.TestFactory3("bye")
    assert z3.value == "bye"

    with pytest.raises(TypeError) as excinfo:
        m.TestFactory3(tag.null_ptr)
    assert str(excinfo.value) == "pybind11::init(): factory function returned nullptr"

    assert [i.alive() for i in cstats] == [3, 3, 3]
    assert ConstructorStats.detail_reg_inst() == n_inst + 9

    del x1, y2, y3, z3
    assert [i.alive() for i in cstats] == [2, 2, 1]
    assert ConstructorStats.detail_reg_inst() == n_inst + 5
    del x2, x3, y1, z1, z2
    assert [i.alive() for i in cstats] == [0, 0, 0]
    assert ConstructorStats.detail_reg_inst() == n_inst

    assert [i.values() for i in cstats] == [
        ["3", "hi!"],
        ["7", "hi again"],
        ["42", "bye"]
    ]
    assert [i.default_constructions for i in cstats] == [1, 1, 1]


def test_init_factory_signature(msg):
    with pytest.raises(TypeError) as excinfo:
        m.TestFactory1("invalid", "constructor", "arguments")
    assert msg(excinfo.value) == """
        __init__(): incompatible constructor arguments. The following argument types are supported:
            1. m.factory_constructors.TestFactory1(arg0: m.factory_constructors.tag.unique_ptr_tag, arg1: int)
            2. m.factory_constructors.TestFactory1(arg0: str)
            3. m.factory_constructors.TestFactory1(arg0: m.factory_constructors.tag.pointer_tag)
            4. m.factory_constructors.TestFactory1(arg0: handle, arg1: int, arg2: handle)

        Invoked with: 'invalid', 'constructor', 'arguments'
    """  # noqa: E501 line too long

    assert msg(m.TestFactory1.__init__.__doc__) == """
        __init__(*args, **kwargs)
        Overloaded function.

        1. __init__(self: m.factory_constructors.TestFactory1, arg0: m.factory_constructors.tag.unique_ptr_tag, arg1: int) -> None

        2. __init__(self: m.factory_constructors.TestFactory1, arg0: str) -> None

        3. __init__(self: m.factory_constructors.TestFactory1, arg0: m.factory_constructors.tag.pointer_tag) -> None

        4. __init__(self: m.factory_constructors.TestFactory1, arg0: handle, arg1: int, arg2: handle) -> None
    """  # noqa: E501 line too long


def test_init_factory_casting():
    """Tests py::init_factory() wrapper with various upcasting and downcasting returns"""

    cstats = [ConstructorStats.get(c) for c in [m.TestFactory3, m.TestFactory4, m.TestFactory5]]
    cstats[0].alive()  # force gc
    n_inst = ConstructorStats.detail_reg_inst()

    # Construction from derived references:
    a = m.TestFactory3(tag.pointer, tag.TF4, 4)
    assert a.value == "4"
    b = m.TestFactory3(tag.shared_ptr, tag.TF4, 5)
    assert b.value == "5"
    c = m.TestFactory3(tag.pointer, tag.TF5, 6)
    assert c.value == "6"
    d = m.TestFactory3(tag.shared_ptr, tag.TF5, 7)
    assert d.value == "7"

    assert ConstructorStats.detail_reg_inst() == n_inst + 4

    # Shared a lambda with TF3:
    e = m.TestFactory4(tag.pointer, tag.TF4, 8)
    assert e.value == "8"

    assert ConstructorStats.detail_reg_inst() == n_inst + 5
    assert [i.alive() for i in cstats] == [5, 3, 2]

    del a
    assert [i.alive() for i in cstats] == [4, 2, 2]
    assert ConstructorStats.detail_reg_inst() == n_inst + 4

    del b, c, e
    assert [i.alive() for i in cstats] == [1, 0, 1]
    assert ConstructorStats.detail_reg_inst() == n_inst + 1

    del d
    assert [i.alive() for i in cstats] == [0, 0, 0]
    assert ConstructorStats.detail_reg_inst() == n_inst

    assert [i.values() for i in cstats] == [
        ["4", "5", "6", "7", "8"],
        ["4", "5", "8"],
        ["6", "7"]
    ]


def test_init_factory_alias():
    """Tests py::init_factory() wrapper with value conversions and alias types"""

    cstats = [m.TestFactory6.get_cstats(), m.TestFactory6.get_alias_cstats()]
    cstats[0].alive()  # force gc
    n_inst = ConstructorStats.detail_reg_inst()

    a = m.TestFactory6(tag.base, 1)
    assert a.get() == 1
    assert not a.has_alias()
    b = m.TestFactory6(tag.alias, "hi there")
    assert b.get() == 8
    assert b.has_alias()
    c = m.TestFactory6(tag.alias, 3)
    assert c.get() == 3
    assert c.has_alias()
    d = m.TestFactory6(tag.alias, tag.pointer, 4)
    assert d.get() == 4
    assert d.has_alias()
    e = m.TestFactory6(tag.base, tag.pointer, 5)
    assert e.get() == 5
    assert not e.has_alias()
    f = m.TestFactory6(tag.base, tag.alias, tag.pointer, 6)
    assert f.get() == 6
    assert f.has_alias()

    assert ConstructorStats.detail_reg_inst() == n_inst + 6
    assert [i.alive() for i in cstats] == [6, 4]

    del a, b, e
    assert [i.alive() for i in cstats] == [3, 3]
    assert ConstructorStats.detail_reg_inst() == n_inst + 3
    del f, c, d
    assert [i.alive() for i in cstats] == [0, 0]
    assert ConstructorStats.detail_reg_inst() == n_inst

    class MyTest(m.TestFactory6):
        def __init__(self, *args):
            m.TestFactory6.__init__(self, *args)

        def get(self):
            return -5 + m.TestFactory6.get(self)

    # Return Class by value, moved into new alias:
    z = MyTest(tag.base, 123)
    assert z.get() == 118
    assert z.has_alias()

    # Return alias by value, moved into new alias:
    y = MyTest(tag.alias, "why hello!")
    assert y.get() == 5
    assert y.has_alias()

    # Return Class by pointer, moved into new alias then original destroyed:
    x = MyTest(tag.base, tag.pointer, 47)
    assert x.get() == 42
    assert x.has_alias()

    assert ConstructorStats.detail_reg_inst() == n_inst + 3
    assert [i.alive() for i in cstats] == [3, 3]
    del x, y, z
    assert [i.alive() for i in cstats] == [0, 0]
    assert ConstructorStats.detail_reg_inst() == n_inst

    assert [i.values() for i in cstats] == [
        ["1", "8", "3", "4", "5", "6", "123", "10", "47"],
        ["hi there", "3", "4", "6", "move", "123", "why hello!", "move", "47"]
    ]


def test_init_factory_dual():
    """Tests init factory functions with dual main/alias factory functions"""
    from pybind11_tests.factory_constructors import TestFactory7

    cstats = [TestFactory7.get_cstats(), TestFactory7.get_alias_cstats()]
    cstats[0].alive()  # force gc
    n_inst = ConstructorStats.detail_reg_inst()

    class PythFactory7(TestFactory7):
        def get(self):
            return 100 + TestFactory7.get(self)

    a1 = TestFactory7(1)
    a2 = PythFactory7(2)
    assert a1.get() == 1
    assert a2.get() == 102
    assert not a1.has_alias()
    assert a2.has_alias()

    b1 = TestFactory7(tag.pointer, 3)
    b2 = PythFactory7(tag.pointer, 4)
    assert b1.get() == 3
    assert b2.get() == 104
    assert not b1.has_alias()
    assert b2.has_alias()

    c1 = TestFactory7(tag.mixed, 5)
    c2 = PythFactory7(tag.mixed, 6)
    assert c1.get() == 5
    assert c2.get() == 106
    assert not c1.has_alias()
    assert c2.has_alias()

    d1 = TestFactory7(tag.base, tag.pointer, 7)
    d2 = PythFactory7(tag.base, tag.pointer, 8)
    assert d1.get() == 7
    assert d2.get() == 108
    assert not d1.has_alias()
    assert d2.has_alias()

    # Both return an alias; the second multiplies the value by 10:
    e1 = TestFactory7(tag.alias, tag.pointer, 9)
    e2 = PythFactory7(tag.alias, tag.pointer, 10)
    assert e1.get() == 9
    assert e2.get() == 200
    assert e1.has_alias()
    assert e2.has_alias()

    f1 = TestFactory7(tag.shared_ptr, tag.base, 11)
    f2 = PythFactory7(tag.shared_ptr, tag.base, 12)
    assert f1.get() == 11
    assert f2.get() == 112
    assert not f1.has_alias()
    assert f2.has_alias()

    g1 = TestFactory7(tag.shared_ptr, tag.invalid_base, 13)
    assert g1.get() == 13
    assert not g1.has_alias()
    with pytest.raises(TypeError) as excinfo:
        PythFactory7(tag.shared_ptr, tag.invalid_base, 14)
    assert (str(excinfo.value) ==
            "pybind11::init(): construction failed: returned holder-wrapped instance is not an "
            "alias instance")

    assert [i.alive() for i in cstats] == [13, 7]
    assert ConstructorStats.detail_reg_inst() == n_inst + 13

    del a1, a2, b1, d1, e1, e2
    assert [i.alive() for i in cstats] == [7, 4]
    assert ConstructorStats.detail_reg_inst() == n_inst + 7
    del b2, c1, c2, d2, f1, f2, g1
    assert [i.alive() for i in cstats] == [0, 0]
    assert ConstructorStats.detail_reg_inst() == n_inst

    assert [i.values() for i in cstats] == [
        ["1", "2", "3", "4", "5", "6", "7", "8", "9", "100", "11", "12", "13", "14"],
        ["2", "4", "6", "8", "9", "100", "12"]
    ]


def test_no_placement_new(capture):
    """Prior to 2.2, `py::init<...>` relied on the type supporting placement
    new; this tests a class without placement new support."""
    with capture:
        a = m.NoPlacementNew(123)

    found = re.search(r'^operator new called, returning (\d+)\n$', str(capture))
    assert found
    assert a.i == 123
    with capture:
        del a
        pytest.gc_collect()
    assert capture == "operator delete called on " + found.group(1)

    with capture:
        b = m.NoPlacementNew()

    found = re.search(r'^operator new called, returning (\d+)\n$', str(capture))
    assert found
    assert b.i == 100
    with capture:
        del b
        pytest.gc_collect()
    assert capture == "operator delete called on " + found.group(1)


def test_multiple_inheritance():
    class MITest(m.TestFactory1, m.TestFactory2):
        def __init__(self):
            m.TestFactory1.__init__(self, tag.unique_ptr, 33)
            m.TestFactory2.__init__(self, tag.move)

    a = MITest()
    assert m.TestFactory1.value.fget(a) == "33"
    assert m.TestFactory2.value.fget(a) == "(empty2)"


def create_and_destroy(*args):
    a = m.NoisyAlloc(*args)
    print("---")
    del a
    pytest.gc_collect()


def strip_comments(s):
    return re.sub(r'\s+#.*', '', s)


def test_reallocations(capture, msg):
    """When the constructor is overloaded, previous overloads can require a preallocated value.
    This test makes sure that such preallocated values only happen when they might be necessary,
    and that they are deallocated properly"""

    pytest.gc_collect()

    with capture:
        create_and_destroy(1)
    assert msg(capture) == """
        noisy new
        noisy placement new
        NoisyAlloc(int 1)
        ---
        ~NoisyAlloc()
        noisy delete
    """
    with capture:
        create_and_destroy(1.5)
    assert msg(capture) == strip_comments("""
        noisy new               # allocation required to attempt first overload
        noisy delete            # have to dealloc before considering factory init overload
        noisy new               # pointer factory calling "new", part 1: allocation
        NoisyAlloc(double 1.5)  # ... part two, invoking constructor
        ---
        ~NoisyAlloc()  # Destructor
        noisy delete   # operator delete
    """)

    with capture:
        create_and_destroy(2, 3)
    assert msg(capture) == strip_comments("""
        noisy new          # pointer factory calling "new", allocation
        NoisyAlloc(int 2)  # constructor
        ---
        ~NoisyAlloc()  # Destructor
        noisy delete   # operator delete
    """)

    with capture:
        create_and_destroy(2.5, 3)
    assert msg(capture) == strip_comments("""
        NoisyAlloc(double 2.5)  # construction (local func variable: operator_new not called)
        noisy new               # return-by-value "new" part 1: allocation
        ~NoisyAlloc()           # moved-away local func variable destruction
        ---
        ~NoisyAlloc()  # Destructor
        noisy delete   # operator delete
    """)

    with capture:
        create_and_destroy(3.5, 4.5)
    assert msg(capture) == strip_comments("""
        noisy new               # preallocation needed before invoking placement-new overload
        noisy placement new     # Placement new
        NoisyAlloc(double 3.5)  # construction
        ---
        ~NoisyAlloc()  # Destructor
        noisy delete   # operator delete
    """)

    with capture:
        create_and_destroy(4, 0.5)
    assert msg(capture) == strip_comments("""
        noisy new          # preallocation needed before invoking placement-new overload
        noisy delete       # deallocation of preallocated storage
        noisy new          # Factory pointer allocation
        NoisyAlloc(int 4)  # factory pointer construction
        ---
        ~NoisyAlloc()  # Destructor
        noisy delete   # operator delete
    """)

    with capture:
        create_and_destroy(5, "hi")
    assert msg(capture) == strip_comments("""
        noisy new            # preallocation needed before invoking first placement new
        noisy delete         # delete before considering new-style constructor
        noisy new            # preallocation for second placement new
        noisy placement new  # Placement new in the second placement new overload
        NoisyAlloc(int 5)    # construction
        ---
        ~NoisyAlloc()  # Destructor
        noisy delete   # operator delete
    """)


@pytest.unsupported_on_py2
def test_invalid_self():
    """Tests invocation of the pybind-registered base class with an invalid `self` argument.  You
    can only actually do this on Python 3: Python 2 raises an exception itself if you try."""
    class NotPybindDerived(object):
        pass

    # Attempts to initialize with an invalid type passed as `self`:
    class BrokenTF1(m.TestFactory1):
        def __init__(self, bad):
            if bad == 1:
                a = m.TestFactory2(tag.pointer, 1)
                m.TestFactory1.__init__(a, tag.pointer)
            elif bad == 2:
                a = NotPybindDerived()
                m.TestFactory1.__init__(a, tag.pointer)

    # Same as above, but for a class with an alias:
    class BrokenTF6(m.TestFactory6):
        def __init__(self, bad):
            if bad == 1:
                a = m.TestFactory2(tag.pointer, 1)
                m.TestFactory6.__init__(a, tag.base, 1)
            elif bad == 2:
                a = m.TestFactory2(tag.pointer, 1)
                m.TestFactory6.__init__(a, tag.alias, 1)
            elif bad == 3:
                m.TestFactory6.__init__(NotPybindDerived.__new__(NotPybindDerived), tag.base, 1)
            elif bad == 4:
                m.TestFactory6.__init__(NotPybindDerived.__new__(NotPybindDerived), tag.alias, 1)

    for arg in (1, 2):
        with pytest.raises(TypeError) as excinfo:
            BrokenTF1(arg)
        assert str(excinfo.value) == "__init__(self, ...) called with invalid `self` argument"

    for arg in (1, 2, 3, 4):
        with pytest.raises(TypeError) as excinfo:
            BrokenTF6(arg)
        assert str(excinfo.value) == "__init__(self, ...) called with invalid `self` argument"



import numpy as np
import numpy.linalg as la

class MaximumIterations(RuntimeError):
  """
    An error to use if an iterative method exceeds the maximum allowed iterations.
  """
  def __init__(self):
    super(MaximumIterations, self).__init__("Exceeded the maximum allowed iterations!")

class MaximumSubdivisions(RuntimeError):
  """
    An error to use if adaptive substepping faisl.
  """
  def __init__(self):
    super(MaximumSubdivisions, self).__init__("Exceeded the maximum allowed step subdivisions!")

def newton_precalc(RJ, x0, verbose = False, rtol = 1.0e-6, atol = 1.0e-10, 
    miter = 50):
  """
    Manually-coded newton-raphson with precalculated jacobian

    This version doesn't do linesearch because ironically that would 
    require re-inverting everything.

    Parameters:
      RJ            function return the residual + a function that gives
                    the action of the inverse jacobian on a vector
      x0            initial guess

    Optional:
      verbose       verbose output
      rtol          relative tolerance
      atol          absolute tolerance
      miter         maximum iterations
  """
  R, J = RJ(x0)
  nR = la.norm(R)
  nR0 = nR
  x = np.copy(x0)
  
  i = 0

  if verbose:
    print("Iter.\tnR\t\tnR/nR0")
    print("%i\t%e\t%e\t" % (i, nR, 1.0))

  while (nR > rtol * nR0) and (nR > atol):
    a = J(R)
    x -= a
    R, J = RJ(x)
    nR = la.norm(R)
    i += 1
    if verbose:
      print("%i\t%e\t%e" % (i, nR, nR / nR0))
    if i > miter:
      if verbose:
        print("")
      raise MaximumIterations()
  
  if verbose:
    print("")

  return x

def newton(RJ, x0, verbose = False, rtol = 1.0e-6, atol = 1.0e-10, miter = 50,
    linesearch = 'none', bt_tau = 0.5, bt_c = 1.0e-4):
  """
    Manually-code newton-raphson so that I can output convergence info, if
    requested.

    Parameters:
      RJ            function return the residual + jacobian
      x0            initial guess

    Optional:
      verbose       verbose output
      rtol          relative tolerance
      atol          absolute tolerance
      miter         maximum iterations
      linesearch    available options: "none" and "backtracking"
      bt_tau        tau factor for backtracking line search
      bt_c          c factor for backtracking line search
  """
  R, J = RJ(x0)
  nR = la.norm(R)
  nR0 = nR
  x = np.copy(x0)
  
  i = 0

  if verbose:
    print("Iter.\tnR\t\tnR/nR0\t\tcond\t\tlinesearch")
    print("%i\t%e\t%e\t" % (i, nR, nR / nR0))

  while (nR > rtol * nR0) and (nR > atol):
    a = la.solve(J, R)

    if linesearch == 'none':
      f = 1.0
    elif linesearch == 'backtracking':
      f = backtrack(RJ, R, J, x, -a, tau = bt_tau, c = bt_c, verbose = verbose)
    else:
      raise ValueError("Unknown linesearch type.")
    
    x -= (a * f)
    R, J = RJ(x)
    nR = la.norm(R)
    i += 1
    if verbose:
      print("%i\t%e\t%e\t%e\t%f" % (i, nR, nR / nR0,la.cond(J), f))
    if i > miter:
      if verbose:
        print("")
      raise MaximumIterations()
  
  if verbose:
    print("")

  return x

def backtrack(RJ, R0, J0, x, a, tau = 0.5, c = 1.0e-4, verbose = False):
  """
    Do a backtracking line search

    Parameters:
      RJ        residual/jacobian function
      R0        value of R, to save a feval
      J0        value of J, to save a feval
      x         point to start from
      a         direction

    Optional:
      tau       backtrack tau
      c         backtrack c
      verbose   verbose output
  """
  alpha = 1.0
  cv = la.norm(RJ(x + alpha * a)[0])
  while cv > la.norm(R0 + c * alpha * np.dot(J0, a)):
    alpha *= tau
    cv = la.norm(RJ(x + alpha * a)[0])

  return alpha

def quasi_newton(RJ, x0, verbose = False, rtol = 1.0e-6, atol = 1.0e-10, 
    miter = 20):
  """
    Manually-code quasi newton-raphson so that I can output convergence info, if
    requested.

    Parameters:
      RJ        function return the residual + jacobian
      x0        initial guess

    Optional:
      verbose   verbose output
      rtol      relative tolerance
      atol      absolute tolerance
      miter     maximum number of iterations
  """
  R, J = RJ(x0)
  J_use = np.copy(J)
  nR = la.norm(R)
  nR0 = nR
  x = np.copy(x0)
  
  i = 0

  if verbose:
    print("Iter.\tnR\t\tnR/nR0\t\tcond")
    print("%i\t%e\t%e\t" % (i, nR, nR / nR0))

  while (nR > rtol * nR0) and (nR > atol):
    x -= la.solve(J_use, R)
    
    R, J = RJ(x)
    nR = la.norm(R)
    i += 1
    if verbose:
      print("%i\t%e\t%e\t%e" % (i, nR, nR / nR0,la.cond(J_use)))
    if i > miter:
      if verbose:
        print("")
      raise MaximumIterations()
  
  if verbose:
    print("")

  return x

def scalar_newton(RJ, x0, verbose = False, rtol = 1.0e-6, atol = 1.0e-10,
    miter = 20, quasi = None):
  """
    Manually-code newton-raphson for scalar problems

    Parameters:
      RJ        function return the residual + jacobian
      x0        initial guess

    Optional:
      verbose   verbose output
      rtol      relative tolerance
      atol      absolute tolerance
      miter     maximum solver iterations
      quasi     if true use a quasi-newton method
  """
  R, J = RJ(x0)
  nR = np.abs(R)
  nR0 = nR
  x = x0
  
  i = 0

  if verbose:
    print("Iter.\tnR\t\tnR/nR0")
    print("%i\t%e\t%e\t" % (i, nR, nR / nR0))

  while (nR > rtol * nR0) and (nR > atol):
    if quasi is not None:
      dx = -R/J
      x += quasi * dx
    else:
      x -= R/J
    R, J = RJ(x)
    nR = np.abs(R)
    i += 1
    if verbose:
      print("%i\t%e\t%e\t" % (i, nR, nR / nR0))
    
    if i > miter:
      if verbose:
        print("")
      raise MaximumIterations()
  
  if verbose:
    print("")

  return x




import sys
sys.path.append('..')

from neml import interpolate, solvers, models, elasticity, ri_flow, hardening, surfaces, visco_flow, general_flow, creep, damage, larsonmiller
from common import *

import unittest
import numpy as np
import numpy.linalg as la

class CommonStandardDamageModel(object):
  """
    Tests that apply to any standard damage model
  """
  def effective(self, s):
    sdev = make_dev(s)
    return np.sqrt(3.0/2.0 * np.dot(sdev, sdev))

  def test_damage(self):
    d_model = self.model.damage(self.d_np1, self.d_n, self.e_np1, self.e_n,
        self.s_np1, self.s_n, self.T_np1, self.T_n, self.t_np1, self.t_n)
    S = self.elastic.S(self.T_np1)
    dS = self.s_np1 - self.s_n
    dee = np.dot(S, dS)
    de = self.e_np1 - self.e_n
    
    dp = np.sqrt(2.0/3.0 * (np.dot(de, de) + np.dot(dee, dee) - 
      2.0 * np.dot(dee, de)))
    f = self.model.f(self.s_np1, self.d_np1, self.T_np1)
    d_calcd = self.d_n + f * dp

    self.assertTrue(np.isclose(d_model, d_calcd))

  def test_function_derivative_s(self):
    d_model = self.model.df_ds(self.stress, self.d_np1, self.T)
    d_calcd = differentiate(lambda x: self.model.f(x, self.d_np1, self.T),
        self.stress)
    self.assertTrue(np.allclose(d_model, d_calcd))

  def test_function_derivative_d(self):
    d_model = self.model.df_dd(self.stress, self.d, self.T)
    d_calcd = differentiate(lambda x: self.model.f(self.s_np1, x, self.T),
        self.d)
    self.assertTrue(np.isclose(d_model, d_calcd))

class CommonScalarDamageModel(object):
  def test_ndamage(self):
    self.assertEqual(self.model.ndamage, 1)

  def test_init_damage(self):
    self.assertTrue(np.allclose(self.model.init_damage(), np.zeros((1,))))

  def test_ddamage_ddamage(self):
    dd_model = self.model.ddamage_dd(self.d_np1, self.d_n, self.e_np1, self.e_n,
        self.s_np1, self.s_n, self.T_np1, self.T_n, self.t_np1, self.t_n)
    dfn = lambda d: self.model.damage(d, self.d_n, self.e_np1, self.e_n,
        self.s_np1, self.s_n, self.T_np1, self.T_n, self.t_np1, self.t_n)
    dd_calcd = differentiate(dfn, self.d_np1)

    print(dd_model, dd_calcd)
    
    self.assertTrue(np.isclose(dd_model, dd_calcd))

  def test_ddamage_dstrain(self):
    dd_model = self.model.ddamage_de(self.d_np1, self.d_n, self.e_np1, self.e_n,
        self.s_np1, self.s_n, self.T_np1, self.T_n, self.t_np1, self.t_n)
    dfn = lambda e: self.model.damage(self.d_np1, self.d_n, e, self.e_n,
        self.s_np1, self.s_n, self.T_np1, self.T_n, self.t_np1, self.t_n)
    dd_calcd = differentiate(dfn, self.e_np1)[0]

    self.assertTrue(np.allclose(dd_model, dd_calcd, rtol = 1.0e-3))

  def test_ddamage_dstress(self):
    dd_model = self.model.ddamage_ds(self.d_np1, self.d_n, self.e_np1, self.e_n,
        self.s_np1, self.s_n, self.T_np1, self.T_n, self.t_np1, self.t_n)
    dfn = lambda s: self.model.damage(self.d_np1, self.d_n, self.e_np1, self.e_n,
        s, self.s_n, self.T_np1, self.T_n, self.t_np1, self.t_n)
    dd_calcd = differentiate(dfn, self.s_np1)[0]

    self.assertTrue(np.allclose(dd_model, dd_calcd, rtol = 1.0e-3))

  def test_nparams(self):
    self.assertEqual(self.model.nparams, 7)

  def test_init_x(self):
    trial_state = self.model.make_trial_state(
        self.e_np1, self.e_n,
        self.T_np1, self.T_n, self.t_np1, self.t_n,
        self.s_n, self.hist_n, self.u_n, self.p_n)
    
    me = np.array(list(self.s_n) + [self.d_n])
    them = self.model.init_x(trial_state)

    self.assertTrue(np.allclose(me, them))

  def test_R(self):
    trial_state = self.model.make_trial_state(
        self.e_np1, self.e_n,
        self.T_np1, self.T_n, self.t_np1, self.t_n,
        self.s_n, self.hist_n, self.u_n, self.p_n)

    R, J = self.model.RJ(self.x_trial, trial_state)

    s_trial = self.x_trial[:6]
    w_trial = self.x_trial[6]

    R_calc = np.zeros((7,))
    s_p_np1, h_p, A_p, u_p, p_p =self.bmodel.update_sd(self.e_np1, self.e_n,
        self.T_np1, self.T_n,
        self.t_np1, self.t_n, self.s_n / (1-self.d_n), self.hist_n[1:],
        self.u_n, self.p_n)
    R_calc[:6] = s_trial - (1-w_trial) * s_p_np1
    d_np1 = self.model.damage(w_trial, self.d_n, self.e_np1, self.e_n, 
        s_trial / (1 - w_trial), self.s_n / (1-self.d_n), self.T_np1,
        self.T_n, self.t_np1, self.t_n)
    R_calc[6] = w_trial - d_np1

    self.assertTrue(np.allclose(R_calc, R))

  def test_jacobian(self):
    trial_state = self.model.make_trial_state(
        self.e_np1, self.e_n,
        self.T_np1, self.T_n, self.t_np1, self.t_n,
        self.s_n, self.hist_n, self.u_n, self.p_n)

    R, J = self.model.RJ(self.x_trial, trial_state)
    Jnum = differentiate(lambda x: self.model.RJ(x, trial_state)[0], 
        self.x_trial)
    
    self.assertTrue(np.allclose(J, Jnum, rtol = 1.0e-3))

class CommonDamagedModel(object):
  def test_nstore(self):
    self.assertEqual(self.model.nstore, self.bmodel.nstore + self.model.ndamage)

  def test_store(self):
    base = self.bmodel.init_store()
    damg = self.model.init_damage()
    comp = list(damg) + list(base)
    fromm = self.model.init_store()
    self.assertTrue(np.allclose(fromm, comp))
  
  def test_tangent_proportional_strain(self):
    t_n = 0.0
    e_n = np.zeros((6,))
    s_n = np.zeros((6,))
    hist_n = self.model.init_store()
    u_n = 0.0
    p_n = 0.0

    for m in np.linspace(0,1,self.nsteps+1)[1:]:
      t_np1 = m * self.ttarget
      e_np1 = m * self.etarget
     
      trial_state = self.model.make_trial_state(
          e_np1, e_n,
          self.T, self.T, t_np1, t_n,
          s_n, hist_n, u_n, p_n)

      s_np1, hist_np1, A_np1, u_np1, p_np1 = self.model.update_sd(
          e_np1, e_n, self.T, self.T, t_np1, t_n, s_n, hist_n,
          u_n, p_n)

      A_num = differentiate(lambda e: self.model.update_sd(e, e_n,
        self.T, self.T, t_np1, t_n, s_n, hist_n, u_n, p_n)[0], e_np1)
      
      print(A_num)
      print(A_np1)
      self.assertTrue(np.allclose(A_num, A_np1, rtol = 5.0e-2, atol = 1.0e-1))

      e_n = np.copy(e_np1)
      s_n = np.copy(s_np1)
      hist_n = np.copy(hist_np1)
      u_n = u_np1
      p_n = p_np1
      t_n = t_np1

class TestWorkDamage(unittest.TestCase, CommonScalarDamageModel,
    CommonDamagedModel):
  def setUp(self):
    self.E = 92000.0
    self.nu = 0.3

    self.s0 = 180.0
    self.Kp = 1000.0
    self.H = 1000.0

    self.elastic = elasticity.IsotropicLinearElasticModel(self.E, "youngs",
        self.nu, "poissons")

    surface = surfaces.IsoKinJ2()
    iso = hardening.LinearIsotropicHardeningRule(self.s0, self.Kp)
    kin = hardening.LinearKinematicHardeningRule(self.H)
    hrule = hardening.CombinedHardeningRule(iso, kin)

    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    self.bmodel = models.SmallStrainRateIndependentPlasticity(self.elastic, 
        flow)
    
    self.fn = interpolate.PolynomialInterpolate([0.1,5.0, 1e-8])
    self.n = 2.1

    self.model = damage.NEMLWorkDamagedModel_sd(
        self.elastic, self.fn, self.n, self.bmodel, verbose = False)

    self.stress = np.array([100,-50.0,300.0,-99,50.0,125.0]) * 0.75
    self.T = 100.0

    self.s_np1 = self.stress
    self.s_n = np.array([-25,150,250,-25,-100,25]) * 0.0

    self.d_np1 = 0.2
    self.d_n = 0.1

    self.e_np1 = np.array([0.1,-0.01,0.15,-0.05,-0.1,0.15]) * 0.75
    self.e_n = np.array([-0.05,0.025,-0.1,0.2,0.11,0.13]) * 0.0

    self.T_np1 = self.T
    self.T_n = 90.0

    self.t_np1 = 5000.0
    self.t_n = 0.9

    self.u_n = 0.0
    self.p_n = 0.0
  
    # This is a rather boring baseline history state to probe, but I can't
    # think of a better way to get a "generic" history from a generic model
    self.hist_n = np.array([self.d_n] + list(self.bmodel.init_store()))
    self.x_trial = np.array([50,-25,150,-150,190,100.0] + [0.41])

    self.nsteps = 10
    self.etarget = np.array([0.1,-0.025,0.02,0.015,-0.02,-0.05])
    self.ttarget = 2.0

    self.ee = np.dot(self.elastic.S(self.T_np1), self.s_np1*(1.0-self.d_np1) - self.s_n*(1.0-self.d_n))
    self.de = self.e_np1 - self.e_n
    self.dp = self.de - self.ee
    self.dt = self.t_np1 - self.t_n
    self.Wdot = np.dot(self.s_np1*(1.0-self.d_np1), self.dp) / self.dt

  def test_definition(self):
    damage = self.model.damage(self.d_np1, self.d_n, self.e_np1, self.e_n,
        self.s_np1, self.s_n, self.T_np1, self.T_n, self.t_np1, self.t_n)
    should = self.d_n + self.n * self.d_np1**((self.n-1)/self.n) * self.Wdot * self.dt / self.fn.value(self.Wdot) 

    self.assertAlmostEqual(damage, should)

class TestClassicalDamage(unittest.TestCase, CommonScalarDamageModel,
    CommonDamagedModel):
  def setUp(self):
    self.E = 92000.0
    self.nu = 0.3

    self.s0 = 180.0
    self.Kp = 1000.0
    self.H = 1000.0

    self.elastic = elasticity.IsotropicLinearElasticModel(self.E, "youngs",
        self.nu, "poissons")

    surface = surfaces.IsoKinJ2()
    iso = hardening.LinearIsotropicHardeningRule(self.s0, self.Kp)
    kin = hardening.LinearKinematicHardeningRule(self.H)
    hrule = hardening.CombinedHardeningRule(iso, kin)

    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    self.bmodel = models.SmallStrainRateIndependentPlasticity(self.elastic, 
        flow)

    self.xi = 0.478
    self.phi = 1.914
    self.A = 10000000.0

    self.model = damage.ClassicalCreepDamageModel_sd(
        self.elastic, 
        self.A, self.xi, self.phi, self.bmodel)

    self.stress = np.array([100,-50.0,300.0,-99,50.0,125.0])
    self.T = 100.0

    self.s_np1 = self.stress
    self.s_n = np.array([-25,150,250,-25,-100,25])

    self.d_np1 = 0.5
    self.d_n = 0.4

    self.e_np1 = np.array([0.1,-0.01,0.15,-0.05,-0.1,0.15])
    self.e_n = np.array([-0.05,0.025,-0.1,0.2,0.11,0.13])

    self.T_np1 = self.T
    self.T_n = 90.0

    self.t_np1 = 1.0
    self.t_n = 0.0

    self.u_n = 0.0
    self.p_n = 0.0
  
    # This is a rather boring baseline history state to probe, but I can't
    # think of a better way to get a "generic" history from a generic model
    self.hist_n = np.array([self.d_n] + list(self.bmodel.init_store()))
    self.x_trial = np.array([50,-25,150,-150,190,100.0] + [0.41])

    self.nsteps = 10
    self.etarget = np.array([0.1,-0.025,0.02,0.015,-0.02,-0.05])
    self.ttarget = 10.0

class CommonEffectiveStress(object):
  def test_derivative(self):
    nd = differentiate(self.es.effective, self.stress)
    d = self.es.deffective(self.stress)
    self.assertTrue(np.allclose(nd, d))

class TestVonMisesEffectiveStress(unittest.TestCase, CommonEffectiveStress):
  def setUp(self):
    self.es = damage.VonMisesEffectiveStress()

    self.stress = np.array([100,-50.0,300.0,-99,50.0,125.0])

  def test_definition(self):
    S = usym(self.stress)
    D = S - np.trace(S) * np.eye(3) / 3.0
    se = np.sqrt(3.0/2 * np.sum(D*D))
    self.assertTrue(np.isclose(se, self.es.effective(self.stress)))

class TestHuddlestonEffectiveStress(unittest.TestCase, CommonEffectiveStress):
  def setUp(self):
    self.b = 0.25
    self.es = damage.HuddlestonEffectiveStress(self.b)

    self.stress = np.array([100,-50.0,300.0,-99,50.0,125.0])

  def test_definition(self):
    S = usym(self.stress)
    s3, s2, s1 = la.eigvalsh(S)

    J1 = s1 + s2 + s3
    sb = np.sqrt(((s1-s2)**2.0+(s2-s3)**2.0+(s3-s1)**2.0)/2.0)
    ss = np.sqrt(s1**2.0 + s2**2.0 + s3**2.0)

    v1 = sb * np.exp(self.b*(J1/ss-1.0))

    I1 = np.trace(S)
    I2 = 0.5*(np.trace(S)**2.0 - np.trace(np.dot(S,S)))
    Sd = S - np.trace(S) * np.eye(3) / 3.0
    I2p = 0.5*(np.trace(Sd)**2.0 - np.trace(np.dot(Sd,Sd)))
    
    J2p = -I2p # Very odd thing here

    sb2 = np.sqrt(3.0*J2p)
    ss2 = np.sqrt(3.0*J2p+I2)

    v2 = sb2 * np.exp(self.b*(I1/ss2-1.0))
    v3 = self.es.effective(self.stress)
    
    self.assertTrue(np.isclose(sb,sb2))
    self.assertTrue(np.isclose(ss,ss2))
    
    self.assertTrue(np.isclose(v1,v2))
    self.assertTrue(np.isclose(v1,v3))

class TestMaxPrincipalEffectiveStress(unittest.TestCase, CommonEffectiveStress):
  def setUp(self):
    self.es = damage.MaxPrincipalEffectiveStress()

    self.stress = np.array([100,-50.0,300.0,-99,50.0,125.0])

  def test_definition(self):
    S = usym(self.stress)
    vals = la.eigvalsh(S)
    se = np.max(vals)
    self.assertTrue(np.isclose(se, self.es.effective(self.stress)))

class TestMaxSeveralEffectiveStress(unittest.TestCase, CommonEffectiveStress):
  def setUp(self):
    self.es = damage.MaxSeveralEffectiveStress([
      damage.VonMisesEffectiveStress(),
      damage.MaxPrincipalEffectiveStress()])

    self.stress = np.array([100,-50.0,300.0,-99,50.0,125.0])

  def test_definition(self):
    S = usym(self.stress)
    vals = la.eigvalsh(S)
    se1 = np.max(vals)

    D = S - np.trace(S) * np.eye(3) / 3.0
    se2 = np.sqrt(3.0/2 * np.sum(D*D))

    self.assertTrue(np.isclose(max(se1,se2), self.es.effective(self.stress)))

class TestSumSeveralEffectiveStress(unittest.TestCase, CommonEffectiveStress):
  def setUp(self):
    self.stresses = [damage.VonMisesEffectiveStress(), 
        damage.MaxPrincipalEffectiveStress()]
    self.weights = [0.37, 1.0-0.37]
    self.es = damage.SumSeveralEffectiveStress(self.stresses, self.weights)

    self.stress = np.array([100,-50.0,300.0,-99,50.0,125.0])

  def test_definition(self):
    base = sum(w * eff.effective(self.stress) for eff, w in zip(self.stresses,
      self.weights))

    self.assertTrue(np.isclose(base, self.es.effective(self.stress)))

class BaseModularDamage(CommonScalarDamageModel, CommonDamagedModel):
  def complete(self):
    self.E = 92000.0
    self.nu = 0.3

    self.s0 = 180.0
    self.Kp = 1000.0
    self.H = 1000.0

    self.elastic = elasticity.IsotropicLinearElasticModel(self.E, "youngs",
        self.nu, "poissons")

    surface = surfaces.IsoKinJ2()
    iso = hardening.LinearIsotropicHardeningRule(self.s0, self.Kp)
    kin = hardening.LinearKinematicHardeningRule(self.H)
    hrule = hardening.CombinedHardeningRule(iso, kin)

    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    self.bmodel = models.SmallStrainRateIndependentPlasticity(self.elastic, 
        flow)

    self.xi = 0.478
    self.phi = 1.914
    self.A = 10000000.0

    self.model = damage.ModularCreepDamageModel_sd(
        self.elastic, 
        self.A, self.xi, self.phi,
        self.effective_model(),
        self.bmodel)

    self.stress = np.array([100,-50.0,300.0,-99,50.0,125.0])
    self.T = 100.0

    self.s_np1 = self.stress
    self.s_n = np.array([-25,150,250,-25,-100,25])

    self.d_np1 = 0.5
    self.d_n = 0.4

    self.e_np1 = np.array([0.1,-0.01,0.15,-0.05,-0.1,0.15])
    self.e_n = np.array([-0.05,0.025,-0.1,0.2,0.11,0.13])

    self.T_np1 = self.T
    self.T_n = 90.0

    self.t_np1 = 1.0
    self.t_n = 0.0

    self.u_n = 0.0
    self.p_n = 0.0
  
    # This is a rather boring baseline history state to probe, but I can't
    # think of a better way to get a "generic" history from a generic model
    self.hist_n = np.array([self.d_n] + list(self.bmodel.init_store()))
    self.x_trial = np.array([50,-25,150,-150,190,100.0] + [0.41])

    self.nsteps = 10
    self.etarget = np.array([0.1,-0.025,0.02,0.015,-0.02,-0.05])
    self.ttarget = 10.0

class TestModularVonMises(unittest.TestCase, BaseModularDamage):
  def setUp(self):
    self.complete()

  def effective_model(self):
    return damage.VonMisesEffectiveStress()

class TestModularMaxPrincipal(unittest.TestCase, BaseModularDamage):
  def setUp(self):
    self.complete()

  def effective_model(self):
    return damage.MaxPrincipalEffectiveStress()

class TestModularMaxSeveral(unittest.TestCase, BaseModularDamage):
  def setUp(self):
    self.complete()

  def effective_model(self):
    return damage.MaxSeveralEffectiveStress([
      damage.VonMisesEffectiveStress(),
      damage.MaxPrincipalEffectiveStress()])

class TestHuddlestonPrincipal(unittest.TestCase, BaseModularDamage):
  def setUp(self):
    self.complete()

  def effective_model(self):
    return damage.HuddlestonEffectiveStress(0.24)

class TestLMDamage(unittest.TestCase, CommonScalarDamageModel, CommonDamagedModel):
  def setUp(self):
    self.E = 92000.0
    self.nu = 0.3

    self.s0 = 180.0
    self.Kp = 1000.0
    self.H = 1000.0

    self.elastic = elasticity.IsotropicLinearElasticModel(self.E, "youngs",
        self.nu, "poissons")

    surface = surfaces.IsoKinJ2()
    iso = hardening.LinearIsotropicHardeningRule(self.s0, self.Kp)
    kin = hardening.LinearKinematicHardeningRule(self.H)
    hrule = hardening.CombinedHardeningRule(iso, kin)

    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    self.bmodel = models.SmallStrainRateIndependentPlasticity(self.elastic, 
        flow)

    self.fn = interpolate.PolynomialInterpolate([-6.653e-9,2.952e-4,-6.197e-1])
    self.C = 32.06

    self.lmr = larsonmiller.LarsonMillerRelation(self.fn, self.C)
    self.effective = damage.VonMisesEffectiveStress()

    self.model = damage.LarsonMillerCreepDamageModel_sd(
        self.elastic, self.lmr, self.effective, self.bmodel)

    self.stress = np.array([100,-50.0,300.0,-99,50.0,125.0])
    self.T = 100.0

    self.s_np1 = self.stress
    self.s_n = np.array([-25,150,250,-25,-100,25])

    self.d_np1 = 0.5
    self.d_n = 0.4

    self.e_np1 = np.array([0.1,-0.01,0.15,-0.05,-0.1,0.15])
    self.e_n = np.array([-0.05,0.025,-0.1,0.2,0.11,0.13])

    self.T_np1 = self.T
    self.T_n = 90.0

    self.t_np1 = 1.0
    self.t_n = 0.0
    self.dt = self.t_np1 - self.t_n

    self.u_n = 0.0
    self.p_n = 0.0
  
    # This is a rather boring baseline history state to probe, but I can't
    # think of a better way to get a "generic" history from a generic model
    self.hist_n = np.array([self.d_n] + list(self.bmodel.init_store()))
    self.x_trial = np.array([50,-25,150,-150,190,100.0] + [0.41])

    self.nsteps = 10
    self.etarget = np.array([0.1,-0.025,0.02,0.015,-0.02,-0.05])
    self.ttarget = 10.0

  def test_damage(self):
    se = self.effective.effective(self.stress)
    tR = self.lmr.tR(se, self.T)
    exact = self.d_n + self.dt / tR
    model = self.model.damage(self.d_np1, self.d_n, 
        self.e_np1, self.e_n, self.s_np1, self.s_n,
        self.T_np1, self.T_n, self.t_np1, self.t_n)
    self.assertTrue(np.isclose(exact, model))

class TestPowerLawDamage(unittest.TestCase, CommonStandardDamageModel, 
    CommonScalarDamageModel, CommonDamagedModel):
  def setUp(self):
    self.E = 92000.0
    self.nu = 0.3

    self.s0 = 180.0
    self.Kp = 1000.0
    self.H = 1000.0

    self.elastic = elasticity.IsotropicLinearElasticModel(self.E, "youngs",
        self.nu, "poissons")

    surface = surfaces.IsoKinJ2()
    iso = hardening.LinearIsotropicHardeningRule(self.s0, self.Kp)
    kin = hardening.LinearKinematicHardeningRule(self.H)
    hrule = hardening.CombinedHardeningRule(iso, kin)

    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    self.bmodel = models.SmallStrainRateIndependentPlasticity(self.elastic, 
        flow)

    self.A = 8.0e-6
    self.a = 2.2

    self.model = damage.NEMLPowerLawDamagedModel_sd(self.elastic, self.A, self.a, 
        self.bmodel)

    self.stress = np.array([100,-50.0,300.0,-99,50.0,125.0])
    self.T = 100.0
    self.d = 0.45

    self.s_np1 = self.stress
    self.s_n = np.array([-25,150,250,-25,-100,25])

    self.d_np1 = 0.5
    self.d_n = 0.4

    self.e_np1 = np.array([0.1,-0.01,0.15,-0.05,-0.1,0.15])
    self.e_n = np.array([-0.05,0.025,-0.1,0.2,0.11,0.13])

    self.T_np1 = self.T
    self.T_n = 90.0

    self.t_np1 = 1.0
    self.t_n = 0.0

    self.u_n = 0.0
    self.p_n = 0.0
  
    # This is a rather boring baseline history state to probe, but I can't
    # think of a better way to get a "generic" history from a generic model
    self.hist_n = np.array([self.d_n] + list(self.bmodel.init_store()))
    self.x_trial = np.array([50,-25,150,-150,190,100.0] + [0.41])

    self.nsteps = 10
    self.etarget = np.array([0.1,-0.025,0.02,0.015,-0.02,-0.05])
    self.ttarget = 10.0

  def test_function(self):
    f_model = self.model.f(self.stress, self.d_np1, self.T)
    f_calcd = self.A * self.effective(self.stress) ** self.a
    self.assertTrue(np.isclose(f_model, f_calcd))

class TestExponentialDamage(unittest.TestCase, CommonStandardDamageModel, 
    CommonScalarDamageModel, CommonDamagedModel):
  def setUp(self):
    self.E = 92000.0
    self.nu = 0.3

    self.s0 = 180.0
    self.Kp = 1000.0
    self.H = 1000.0

    self.elastic = elasticity.IsotropicLinearElasticModel(self.E,
        "youngs", self.nu, "poissons")

    surface = surfaces.IsoKinJ2()
    iso = hardening.LinearIsotropicHardeningRule(self.s0, self.Kp)
    kin = hardening.LinearKinematicHardeningRule(self.H)
    hrule = hardening.CombinedHardeningRule(iso, kin)

    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    self.bmodel = models.SmallStrainRateIndependentPlasticity(self.elastic, 
        flow)

    self.W0 = 10.0
    self.k0 = 0.0001
    self.a = 2.0

    self.model = damage.NEMLExponentialWorkDamagedModel_sd(
        self.elastic, self.W0, self.k0, 
        self.a, self.bmodel)

    self.stress = np.array([100,-50.0,300.0,-99,50.0,125.0])
    self.T = 100.0
    self.d = 0.45

    self.s_np1 = self.stress
    self.s_n = np.array([-25,150,250,-25,-100,25])

    self.d_np1 = 0.5
    self.d_n = 0.4

    self.e_np1 = np.array([0.1,-0.01,0.15,-0.05,-0.1,0.15])
    self.e_n = np.array([-0.05,0.025,-0.1,0.2,0.11,0.13])

    self.T_np1 = self.T
    self.T_n = 90.0

    self.t_np1 = 1.0
    self.t_n = 0.0

    self.u_n = 0.0
    self.p_n = 0.0
  
    # This is a rather boring baseline history state to probe, but I can't
    # think of a better way to get a "generic" history from a generic model
    self.hist_n = np.array([self.d_n] + list(self.bmodel.init_store()))
    self.x_trial = np.array([50,-25,150,-150,190,100.0] + [0.41])

    self.nsteps = 10
    self.etarget = np.array([0.1,-0.025,0.02,0.015,-0.02,-0.05])
    self.ttarget = 10.0

  def test_function(self):
    f_model = self.model.f(self.stress, self.d_np1, self.T)
    f_calcd = (self.d_np1 + self.k0) ** self.a * self.effective(self.stress) / self.W0

    self.assertTrue(np.isclose(f_model, f_calcd))


class TestCombinedDamage(unittest.TestCase, CommonScalarDamageModel,
    CommonDamagedModel):
  def setUp(self):
    self.E = 92000.0
    self.nu = 0.3

    self.s0 = 180.0
    self.Kp = 1000.0
    self.H = 1000.0

    self.elastic = elasticity.IsotropicLinearElasticModel(self.E,
        "youngs", self.nu, "poissons")

    surface = surfaces.IsoKinJ2()
    iso = hardening.LinearIsotropicHardeningRule(self.s0, self.Kp)
    kin = hardening.LinearKinematicHardeningRule(self.H)
    hrule = hardening.CombinedHardeningRule(iso, kin)

    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    self.bmodel = models.SmallStrainRateIndependentPlasticity(self.elastic, 
        flow)

    self.W0 = 10.0
    self.k0 = 0.0001
    self.a0 = 2.0

    self.model1 = damage.NEMLExponentialWorkDamagedModel_sd(
        self.elastic, self.W0, self.k0, 
        self.a0, self.bmodel)

    self.W02 = 10.0
    self.k02 = 0.001
    self.a02 = 1.5

    self.model2 = damage.NEMLExponentialWorkDamagedModel_sd(
        self.elastic, self.W02, self.k02, 
        self.a02, self.bmodel)

    self.model = damage.CombinedDamageModel_sd(self.elastic, 
        [self.model1, self.model2], self.bmodel)

    self.stress = np.array([100,-50.0,300.0,-99,50.0,125.0])
    self.T = 100.0
    self.d = 0.45

    self.s_np1 = self.stress
    self.s_n = np.array([-25,150,250,-25,-100,25])

    self.d_np1 = 0.5
    self.d_n = 0.4

    self.e_np1 = np.array([0.1,-0.01,0.15,-0.05,-0.1,0.15])
    self.e_n = np.array([-0.05,0.025,-0.1,0.2,0.11,0.13])

    self.T_np1 = self.T
    self.T_n = 90.0

    self.t_np1 = 1.0
    self.t_n = 0.0

    self.u_n = 0.0
    self.p_n = 0.0
  
    # This is a rather boring baseline history state to probe, but I can't
    # think of a better way to get a "generic" history from a generic model
    self.hist_n = np.array([self.d_n] + list(self.bmodel.init_store()))
    self.x_trial = np.array([50,-25,150,-150,190,100.0] + [0.41])

    self.nsteps = 10
    self.etarget = np.array([0.1,-0.025,0.02,0.015,-0.02,-0.05])
    self.ttarget = 10.0
